'''
Created on 3 dec. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from zoneinfo import ZoneInfo

from pandas import read_csv
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class PicopattReader(AbstractReader):
    '''
    classdocs
    '''
    COLUMNS = ["project_id", "track_id", "section_duration",
               "section_speed", "section_warning", "section_weather",
               "lon_rtk", "lat_rtk"]

    def __init__(self, inputFile, tzinfo="Europe/Paris", ocrs="epsg:2154"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning
        self.inputFile = inputFile
        self.tzinfo = ZoneInfo(tzinfo)
        self.ocrs = ocrs

    def __get_version(self):
        df = read_csv(PicopattReader.opener(self.inputFile), sep=";", nrows=0)

        if ((37 == len(df.columns)) and
                all([column in df for column in PicopattReader.COLUMNS])):
            return 1.0
        elif ((38 == len(df.columns)) and
                all([column in df for column in PicopattReader.COLUMNS]) and
                ("gnss_accuracy" in df)):
            return 2.0
        raise Exception("File format not as expected!")

    def __fillna(self, df):
        from pandas import DataFrame, isna

        for column in ["project_id", "track_id"]:
            df[column] = df[column].ffill()

        rows, ref = [], {}
        for _, row in df.iterrows():
            sid = row.section_id
            sduration = row.section_duration
            sspeed = row.section_speed
            swarning = row.section_warning
            sweather = row.section_weather

            if isna([sduration, sspeed, swarning, sweather]).all():
                if sid in ref:
                    row.section_duration = ref[sid]["section_duration"]
                    row.section_speed = ref[sid]["section_speed"]
                    row.section_warning = ref[sid]["section_warning"]
                    row.section_weather = ref[sid]["section_weather"]
                else:
                    raise Exception("Unreachable instruction!")
            else:
                ref[sid] = {
                    "section_duration": sduration,
                    "section_speed": sspeed,
                    "section_warning": swarning,
                    "section_weather": sweather
                }
            rows.append(row)
        return DataFrame(rows)

    def __geodataframe(self, df):
        from geopandas import GeoDataFrame
        from shapely import Point

        df["geometry"] = df.apply(
            lambda row: Point(row.lon_rtk, row.lat_rtk), axis=1)
        gdf = GeoDataFrame(df, geometry="geometry", crs="epsg:4326")
        return gdf.to_crs(self.ocrs)

    def run(self):
        version = self.__get_version()

        if (version not in [1.0, 2.0]):
            warnings.warn(
                "Picopatt file format is expected to be equal to 1.0 or 2.0!")

        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        df = read_csv(PicopattReader.opener(self.inputFile),
                      parse_dates=["timestamp"], sep=";")
        # df.timestamp = df.timestamp.apply(lambda dt: dt.replace(tzinfo=self.tzinfo))
        # df.timestamp = df.timestamp.dt.tz_localize(self.tzinfo).dt.tz_convert("UTC")
        df.timestamp = df.timestamp.dt.tz_localize(self.tzinfo)
        df = self.__fillna(df)
        gdf = self.__geodataframe(df)

        if not gdf.point_id.is_monotonic_increasing:
            msg = f"point_id in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        if not gdf.timestamp.is_monotonic_increasing:
            msg = f"Timestamps in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        return gdf


"""
inputFile = "/media/tleduc/disk_20241128/uclimweb/data/picopatt_montpellier/picopatt_montpellier*antigone*20240710*1412.csv"
df = PicopattReader(inputFile).run()
df.to_csv("/tmp/1.csv", sep=";", index=False)
print(df.head(3))

inputFile = "/media/tleduc/disk_20241128/uclimweb/data/picopatt_nantes/picopatt_nantes*loire*20240919*1829.csv"
df = PicopattReader(inputFile).run()
df.to_csv("/tmp/2.csv", sep=";", index=False)
print(df.head(3))
"""



'''
Created on 17 oct. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import clip, GeoDataFrame
from io import StringIO
from pandas import read_csv
from requests import get
from shapely.wkt import loads
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class BDTopoWFSReader(GeoProcess):
    '''
    classdocs
    '''
    URL = "https://wxs.ign.fr/topographie/geoportail/wfs"

    def __init__(self, roi, layer="batiment", clip=False, crs="EPSG:2154"):
        '''
        Constructor
        '''
        if not isinstance(roi, GeoDataFrame):
            raise IllegalArgumentTypeException(roi, "GeoDataFrame")

        self.roi = roi
        self.layer = layer
        self.clip = clip
        self.crs = crs

    @staticmethod
    def buildings(roi, clip=True, crs="EPSG:2154"):
        return BDTopoWFSReader(roi, layer="batiment", clip=clip, crs=crs).run()

    @staticmethod
    def listlayers():
        # https://geoservices.ign.fr/services-web-experts-topographie
        import re

        RE = re.compile(r"BDTOPO_V3:\w*")

        params = dict(
            request="GetCapabilities",
            service="WFS",
            version="2.0.0",
        )

        response = get(BDTopoWFSReader.URL, params=params)
        if (200 == response.status_code):
            html = response.content.decode(encoding="utf-8")
            html = RE.findall(html)
            html = sorted(list(set(html)))
            return html
        raise Exception(f"WFS error: {response.status_code}")

    def __get_features(self):
        minx, miny, maxx, maxy = self.roi.to_crs("EPSG:4326").total_bounds
        params = dict(
            # bbox="{0},{1},{2},{3}".format(*roi.total_bounds),
            bbox="{0},{1},{2},{3}".format(
                miny, minx, maxy, maxx),  # WARNING YX!
            # count=50,
            # filter="",
            # outputFormat="SHAPE-ZIP",
            outputFormat="csv",
            request="GetFeature",
            service="WFS",
            # startIndex=0,
            typeName=f"BDTOPO_V3:{self.layer}",
            version="2.0.0",
            srsname=self.crs,
            # srsname="urn:ogc:def:crs:EPSG::4326",
        )
        response = get(BDTopoWFSReader.URL, params=params)
        if (200 == response.status_code):
            csv = response.content.decode(encoding="utf-8")
            rows = StringIO(csv)
            df = read_csv(rows, decimal=".", sep=",")
            df.rename(columns={"geometrie": "geometry"}, inplace=True)
            df.geometry = df.geometry.apply(lambda g: loads(g))
            return GeoDataFrame(df, crs=self.crs)
        raise Exception(f"WFS error: {response.status_code}")

    @staticmethod
    def roads(roi, clip=True, crs="EPSG:2154"):
        return BDTopoWFSReader(roi, layer="troncon_de_route", clip=clip, crs=crs).run()

    def run(self):
        gdf = self.__get_features()
        if self.clip:
            gdf = clip(gdf, self.roi.total_bounds, keep_geom_type=True)
        return gdf


"""
import matplotlib.pyplot as plt
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos

roi = GeoDataFrameDemos.ensaNantesBuildings()
buildings = BDTopoWFSReader.buildings(roi)
roads = BDTopoWFSReader.roads(roi)

# buildings = BDTopoWFSReader(roi, layer="batiment").run()
# roads = BDTopoWFSReader(roi, layer="troncon_de_route", clip=True).run()

fig, ax = plt.subplots(figsize=(8.26, 8.26))
roi.boundary.plot(ax=ax, color="red")
buildings.plot(ax=ax, color="grey")
roads.plot(ax=ax, color="black")
plt.show()

print(BDTopoWFSReader.listlayers())
"""



'''
Created on 24 sep. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime
from numpy import isnan
from pandas import DataFrame, concat
from pymap3d.aer import ecef2aer
from pytz import timezone
from shapely import Point
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.io.CSVReader import CSVReader
from t4gpd.resilientgaia.SatelliteLib import SatelliteLib


class STECEF2AERSatelliteReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, ifile, lat="lat", lon="lon", alt="alt", sep=",", decimal=".",
                 iepsg="epsg:4326", oepsg="epsg:2154", timestampFieldName=None,
                 tzinfo=timezone("Europe/Paris")):
        '''
        Constructor
        '''
        self.sensors = CSVReader(ifile, xFieldName=lon, yFieldName=lat,
                                 fieldSep=sep, srcEpsgCode=iepsg, dstEpsgCode=oepsg,
                                 decimalSep=decimal).run()
        self.lat, self.lon, self.alt = lat, lon, alt
        self.timestampFieldName = timestampFieldName
        self.tzinfo = tzinfo

    def __ecef2aer(self, row, h0, ell=SatelliteLib.WGS84):
        # ECEF: earth-centered, earth-fixed frame
        # AER: azimuth, elevation, slant range
        # ECEF -> AER
        vp, lat0, lon0 = row.geometry, row[self.lat], row[self.lon]

        nsat, rows = 0, []
        for i in range(SatelliteLib.NSAT):
            x = row[f"satcoordX_{i}"]
            y = row[f"satcoordY_{i}"]
            z = row[f"satcoordZ_{i}"]

            az, el, srange, satpos = None, None, None, Point()
            if not (isnan(x) or isnan(y) or isnan(z)):
                nsat += 1
                # https://geospace-code.github.io/pymap3d/aer.html#pymap3d.aer.ecef2aer
                az, el, srange = ecef2aer(
                    x, y, z, lat0, lon0, h0, ell=ell, deg=True)
            rows.extend([az, el, srange])

        rows.append(nsat)
        return rows

    @staticmethod
    def __get_colname(i):
        if (3 * SatelliteLib.NSAT == i):
            return "nsat"
        match i % 3:
            case 0:
                return f"sat_{i//3}_az"
            case 1:
                return f"sat_{i//3}_el"
            case 2:
                return f"sat_{i//3}_sr"
            case 3:
                return f"sat_{i//3}"

    @staticmethod
    def rename_satellites(sat):
        columns = {}
        for i in range(SatelliteLib.NSAT):
            satName = SatelliteLib.get_satellite_name(i)
            columns.update({
                f"sat_{i}_az": f"{satName}_az",
                f"sat_{i}_el": f"{satName}_el",
                f"sat_{i}_sr": f"{satName}_sr",
            })
        return sat.rename(columns=columns)

    @staticmethod
    def polarPlot(sat, gid, constellations=["GPS", "Galileo", "GLONASS"], title=None, odir=None):
        import matplotlib.pyplot as plt
        from numpy import degrees, linspace, pi, radians

        colors = {"GPS": "green", "Galileo": "blue", "GLONASS": "red"}

        df = STECEF2AERSatelliteReader.rename_satellites(sat)
        df = df.loc[df[df.gid == gid].index]
        df.dropna(axis=1, how="all", inplace=True)

        azims = df.loc[:, [c for c in df.columns if c.endswith("_az")]]
        elevs = df.loc[:, [c for c in df.columns if c.endswith("_el")]]
        labels = [c[:-3] for c in azims.columns]
        constellation = [SatelliteLib.constellation(label) for label in labels]

        azims = radians(azims.to_numpy()).reshape(-1)
        elevs = elevs.to_numpy().reshape(-1)

        df = DataFrame({"constellation": constellation, "labels": labels,
                        "azims": azims, "elevs": elevs
                        })
        groups = df.groupby(by="constellation")

        fig, ax = plt.subplots(
            figsize=(1.4 * 8.26, 2 * 8.26), subplot_kw={"projection": "polar"})
        if title is not None:
            ax.set_title(title, fontsize=24)
        # Set the offset for the location of 0 in radians
        # ax.set_theta_offset(0)
        # clockwise -1, counterclockwise 1
        ax.set_theta_direction(-1)
        # Set the location of theta's zero
        ax.set_theta_zero_location(loc="N")
        ax.set_rlim(bottom=0, top=90)
        ax.set_xticks(ticks=radians(linspace(0, 360, 12, endpoint=False)))
        ax.set_yticks(ticks=range(0, 90, 15))
        ax.invert_yaxis()
        ax.tick_params(axis="both", colors="lightgrey",
                       grid_color="lightgrey", grid_linewidth=1,
                       grid_linestyle="dotted", which="both")
        # ax.set_thetamin(45)
        # ax.set_thetamax(135)
        for constellation, _df in groups:
            if (constellations is None) or (constellation in constellations):
                ax.scatter(_df.azims, _df.elevs, color=colors[constellation],
                           label=constellation, marker="o")
                for azim, elev, label in zip(_df.azims, _df.elevs, _df.labels):
                    ax.text(azim, elev, label, color=colors[constellation])
        ax.legend()
        if odir is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(f"{odir}/polarplot_{gid}.pdf", bbox_inches="tight")
        plt.close(fig)

    """
    @staticmethod
    def polarPlotTraj(sat, constellations=["GPS", "Galileo", "GLONASS"], title=None, odir=None):
        import matplotlib.pyplot as plt
        from numpy import degrees, linspace, pi, radians

        colors = {"GPS": "green", "Galileo": "blue", "GLONASS": "red"}

        df = STECEF2AERSatelliteReader.rename_satellites(sat)
        df.dropna(axis=1, how="all", inplace=True)

        azims = df.loc[:, [c for c in df.columns if c.endswith("_az")]]
        elevs = df.loc[:, [c for c in df.columns if c.endswith("_el")]]
        # labels = [c[:-3] for c in azims.columns]
        # constellation = [SatelliteLib.constellation(label[0])
        #                  for label in labels]

        # azims = radians(azims.to_numpy()).reshape(-1)
        # elevs = elevs.to_numpy().reshape(-1)
        # return df, constellation, labels, azims, elevs
        # df = DataFrame({"constellation": constellation, "labels": labels,
        #                 "azims": azims, "elevs": elevs
        #                 })
        # groups = df.groupby(by="constellation")

        fig, ax = plt.subplots(
            figsize=(1.4 * 8.26, 2 * 8.26), subplot_kw={"projection": "polar"})
        if title is not None:
            ax.set_title(title, fontsize=24)
        # Set the offset for the location of 0 in radians
        # ax.set_theta_offset(0)
        # clockwise -1, counterclockwise 1
        ax.set_theta_direction(-1)
        # Set the location of theta's zero
        ax.set_theta_zero_location(loc="N")
        ax.set_rlim(bottom=0, top=90)
        ax.set_xticks(ticks=radians(linspace(0, 360, 12, endpoint=False)))
        ax.set_yticks(ticks=range(0, 90, 15))
        ax.invert_yaxis()
        ax.tick_params(axis="both", colors="lightgrey",
                       grid_color="lightgrey", grid_linewidth=1,
                       grid_linestyle="dotted", which="both")
        # ax.set_thetamin(45)
        # ax.set_thetamax(135)
        for i in range(SatelliteLib.NSAT):
        # for i in [7]:
            colname = f"E{i:02d}_az"
            if colname in azims:
                _azims = azims.loc[:, colname]
                _elevs = elevs.loc[:, f"E{i:02d}_el"]
                constellation = "GPS"
                constellation = "Galileo"
                ax.scatter(_azims, _elevs, color=colors[constellation],
                           label=constellation, marker="o")
                # for azim, elev, label in zip(_azims, _elevs, _labels):
                #     ax.text(azim, elev, label, color=colors[constellation])


        # for constellation, _df in groups:
        #     if (constellations is None) or (constellation in constellations):
        #         ax.scatter(_df.azims, _df.elevs, color=colors[constellation],
        #                    label=constellation, marker="o")
        #         for azim, elev, label in zip(_df.azims, _df.elevs, _df.labels):
        #             ax.text(azim, elev, label, color=colors[constellation])
        ax.legend()
        if odir is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(f"{odir}/polarplot_{gid}.png", bbox_inches="tight")
        plt.close(fig)
    """

    def run(self):
        h0 = self.sensors.loc[0, self.alt]
        df = self.sensors.apply(lambda row: self.__ecef2aer(row, h0), axis=1)
        df = df.to_frame(name="satellites")
        df = DataFrame(df.satellites.to_list(), columns=[
                       STECEF2AERSatelliteReader.__get_colname(i)
                       for i in range(1+3*SatelliteLib.NSAT)])
        df = concat([self.sensors, df], axis=1)

        if not self.timestampFieldName is None:
            df[self.timestampFieldName] = df[self.timestampFieldName].apply(
                lambda v: datetime.fromtimestamp(v, self.tzinfo))

        return df



'''
Created on 13 mai 2022

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from pandas import read_csv
from shapely import LineString
from shapely.wkt import loads


class GeoDataFrameLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def getBBoxDiagonal(gdf):
        xmin, ymin, xmax, ymax = gdf.total_bounds
        return LineString([(xmin, ymin), (xmax, ymax)])

    @staticmethod
    def isAGeoDataFrameOfBipoints(gdf):
        return (isinstance(gdf, GeoDataFrame) and
                all(gdf.geometry.apply(lambda g: isinstance(g, LineString))) and
                all(gdf.geometry.apply(lambda g: 2 == len(g.coords)))
                )

    @staticmethod
    def read_csv(inputFile, decimal=".", sep=";", crs="epsg:2154"):
        df = read_csv(inputFile, decimal=decimal, sep=sep)
        df.geometry = df.geometry.apply(lambda g: loads(g))
        return GeoDataFrame(df, crs=crs)

    @staticmethod
    def shareTheSameCrs(*gdfs):
        return (
            (0 == len(gdfs)) or
            all([
                (isinstance(gdf, GeoDataFrame) and (gdfs[0].crs == gdf.crs))
                for gdf in gdfs
            ]) or
            all([
                (isinstance(gdf, GeoDataFrame) and (gdf.crs is None))
                for gdf in gdfs
            ])
        )



'''
Created on 21 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import isnan
from pandas import concat
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STSatelliteLabelling(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, skymaps, prjsat, gid="gid"):
        '''
        Constructor
        '''
        if not isinstance(skymaps, GeoDataFrame):
            raise IllegalArgumentTypeException(
                skymaps, "skymaps GeoDataFrame")
        if not gid in skymaps:
            raise Exception(
                f"{gid} is not a relevant field name for skymaps!")
        if not isinstance(prjsat, GeoDataFrame):
            raise IllegalArgumentTypeException(
                prjsat, "prjsat GeoDataFrame")
        if not "gid" in prjsat:
            raise Exception(
                f"{gid} is not a relevant field name for prjsat!")

        if not GeoDataFrameLib.shareTheSameCrs(skymaps, prjsat):
            raise Exception(
                "Illegal argument: skymaps and prjsat must share shames CRS!")

        self.skymaps = skymaps
        self.prjsat = prjsat
        self.gid = gid
        self.gids = self.skymaps[self.gid]

    def __merge(self):
        prjsat = []
        for gid in self.gids:
            _skymap = self.skymaps.loc[self.skymaps[self.skymaps[self.gid] == gid].index, [
                "geometry"]]
            _prjsat = self.prjsat.loc[self.prjsat[self.prjsat[self.gid] == gid].index]
            _prjsat = _prjsat.sjoin(_skymap, how="left")
            prjsat.append(_prjsat)

        prjsat = concat(prjsat)
        prjsat["label"] = prjsat.index_right.apply(
            lambda v: "LOS" if isnan(v) else "NLOS")
        prjsat.drop(columns=["index_right"], inplace=True)
        return prjsat

    def run(self):
        prjsat = self.__merge()
        return prjsat

    @staticmethod
    def test():
        import matplotlib.pyplot as plt
        from datetime import timezone
        from geopandas import read_file
        from matplotlib.colors import ListedColormap
        from t4gpd.resilientgaia.STECEF2AERSatelliteReader import STECEF2AERSatelliteReader
        from t4gpd.resilientgaia.STSatelliteOnSiteProjection import STSatelliteOnSiteProjection
        from t4gpd.skymap.STSkyMap25D import STSkyMap25D

        # LOAD DATASETS
        rootdir = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data"
        ifile = f"{rootdir}/20240917_ReSilientGAIA.gpkg"
        buildings = read_file(ifile, layer="buildings")

        ifile = f"{rootdir}/20240917_parcours_timestamp.csv"
        aersat = STECEF2AERSatelliteReader(
            ifile, timestampFieldName="timeUTC").run()
        aersat["gid"] = range(len(aersat))
        aersat = aersat.set_index("timeUTC").tz_convert(
            timezone.utc).reset_index()

        # gids = [959, 1615, 1718]
        gids = [959]
        aersat = aersat.query(f"gid in {gids}", inplace=False)

        # PROCESS DATASETS
        projectionName, size = "Stereographic", 4.0
        skymaps = STSkyMap25D(buildings, aersat, nRays=64, rayLength=100.0,
                              elevationFieldname="HAUTEUR", h0=0.0, size=size, epsilon=1e-2,
                              projectionName=projectionName, withIndices=True, withAngles=False,
                              encode=True, threshold=1e-6).run()
        prjsat = STSatelliteOnSiteProjection(aersat, gid="gid", timestamp="timeUTC",
                                             proj=projectionName, size=size).run()

        prjsat2 = STSatelliteLabelling(skymaps, prjsat, "gid").run()

        # MAPPING
        my_cmap = ListedColormap(["green", "red"])

        for gid in gids:
            _prjsat2 = prjsat2.query(f"gid == {gid}", inplace=False)
            ht = _prjsat2.groupby(by="label").agg(
                {'constellation': "count"}).to_dict()["constellation"]

            minx, miny, maxx, maxy = skymaps.query(
                f"gid == {gid}").buffer(1.0).total_bounds

            fig, ax = plt.subplots(figsize=(1.15 * 8.26, 1.15 * 8.26))
            ax.set_title(
                f"Position {gid} ({ht['LOS']} LOS + {ht['NLOS']} NLOS)", fontsize=20)
            skymaps.plot(ax=ax, color="lightgrey")
            aersat.plot(ax=ax, color="black", marker="P")
            _prjsat2.plot(ax=ax, column="label", cmap=my_cmap, marker="o")

            ax.axis("off")
            ax.axis([minx, maxx, miny, maxy])
            # plt.savefig(f"satmap_{gid}.pdf", bbox_inches="tight")
            fig.tight_layout()
            plt.show()
            plt.close(fig)

        return prjsat2


# prjsat2 = STSatelliteLabelling.test()

"""
import matplotlib.pyplot as plt
from datetime import timezone
from geopandas import read_file
from matplotlib.colors import ListedColormap
from t4gpd.resilientgaia.STECEF2AERSatelliteReader import STECEF2AERSatelliteReader
from t4gpd.resilientgaia.STSatelliteLabelling import STSatelliteLabelling
from t4gpd.resilientgaia.STSatelliteOnSiteProjection import STSatelliteOnSiteProjection
from t4gpd.skymap.STSkyMap25D import STSkyMap25D

# LOAD DATASETS
ifile = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data/20240917_ReSilientGAIA.gpkg"
buildings = read_file(ifile, layer="buildings")

ifile = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data/20240917_parcours_timestamp.csv"
aersat = STECEF2AERSatelliteReader(ifile, timestampFieldName="timeUTC").run()
aersat["gid"] = range(len(aersat))
aersat = aersat.set_index("timeUTC").tz_convert(timezone.utc).reset_index()
gids = [959, 1615, 1718]
# gids = [959]
aersat = aersat.loc[aersat[aersat.gid.isin(gids)].index]

# PROCESS DATASETS
projectionName, size = "Stereographic", 4.0
skymaps = STSkyMap25D(buildings, aersat, nRays=64, rayLength=100.0,
                      elevationFieldname="HAUTEUR", h0=0.0, size=size, epsilon=1e-2,
                      projectionName=projectionName, withIndices=True, withAngles=False,
                      encode=True, threshold=1e-6).run()
prjsat = STSatelliteOnSiteProjection(aersat, gid="gid", timestamp="timeUTC",
                                     proj=projectionName, size=size).run()
prjsat2 = STSatelliteLabelling(skymaps, prjsat, "gid").run()
"""



'''
Created on 22 feb. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import dtype, floating, integer, issubdtype
from pandas import Series


class TypeLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def are_both_floating_or_integer(data_type1, data_type2):
        return (
            (TypeLib.is_floating(data_type1) and TypeLib.is_floating(data_type2)) or
            (TypeLib.is_integer(data_type1) and TypeLib.is_integer(data_type2))
        )

    @staticmethod
    def is_floating(data_type):
        if not isinstance(data_type, (dtype, Series, type)):
            data_type = type(data_type)
        return issubdtype(data_type, floating)

    @staticmethod
    def is_integer(data_type):
        if not isinstance(data_type, (dtype, Series, type)):
            data_type = type(data_type)
        return issubdtype(data_type, integer)



'''
Created on 21 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timedelta, timezone
from geopandas import GeoDataFrame
from numpy import cos, dot, ndarray
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class SkyMapRadiationBalance(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, skymaps, dt, meteo=None, svfFieldname="svf",
                 anglesFieldname="angles", model="pysolar", encode=False):
        '''
        Constructor
        '''
        if not isinstance(skymaps, GeoDataFrame):
            raise IllegalArgumentTypeException(skymaps, "GeoDataFrame")
        self.skymaps = skymaps

        if not svfFieldname in skymaps:
            raise Exception(f"{svfFieldname} is not a relevant field name!")
        self.svf = svfFieldname

        if not anglesFieldname in skymaps:
            raise Exception(f"{anglesFieldname} is not a relevant field name!")
        self.angles = anglesFieldname

        sunModel = SunLib(gdf=skymaps, model=model)
        if isinstance(dt, datetime):
            day = date(*dt.timetuple()[0:3])
            dt0 = sunModel.getSunrise(day) + timedelta(hours=1)
            dt1 = dt
        elif isinstance(dt, date):
            dt0 = sunModel.getSunrise(dt) + timedelta(hours=1)
            dt1 = sunModel.getSunset(dt)
        elif (isinstance(dt, (list, ndarray, tuple)) and
              isinstance(dt[0], datetime) and isinstance(dt[1], datetime)):
            dt0, dt1 = dt
            if (date(*dt0.timetuple()[0:3]) != date(*dt1.timetuple()[0:3])):
                raise IllegalArgumentTypeException(
                    dt, "Pair of datetimes must occur on the same day!")
            if (dt0 > dt1):
                raise IllegalArgumentTypeException(
                    dt, "Pair of datetimes must be ordered!")
        else:
            raise IllegalArgumentTypeException(
                dt, "single date or datetime; or pair of datetimes")

        self.dts = [
            datetime(*dt0.timetuple()[0:3], h, tzinfo=timezone.utc)
            for h in range(dt0.hour, dt1.hour+1)
        ]

        if meteo is None:
            day = date(*self.dts[0].timetuple()[0:3])
            self.irrad = SkyMapRadiationBalance.__theoretical_irrad(
                sunModel, day)
        else:
            if "timestamp" in meteo:
                day = date(*self.dts[0].timetuple()[0:3])
                self.irrad = SkyMapRadiationBalance.__irrad_from_daily_meteo(
                    meteo, day)
            else:
                month = self.dts[0].month
                self.irrad = SkyMapRadiationBalance.__irrad_from_monthly_meteo(
                    meteo, month)
        self.encode = encode

    @staticmethod
    def __irrad_from_daily_meteo(meteo, day):
        cols = ["hour", "sun_beam_dir", "solar_alti",
                "solar_azim", "GHI", "DNI", "DHI",]
        _day = day.strftime("%Y%m%d")
        _meteo = meteo.loc[meteo[
            meteo.timestamp.apply(lambda dt: dt.strftime("%Y%m%d") == _day)
        ].index, cols]
        _meteo.set_index("hour", drop=True, inplace=True)
        return _meteo.to_dict(orient="index")

    @staticmethod
    def __irrad_from_monthly_meteo(meteo, month):
        cols = ["hour", "sun_beam_dir", "solar_alti",
                "solar_azim", "GHI", "DNI", "DHI"]
        _meteo = meteo.loc[meteo[meteo.month == month].index, cols]
        _meteo.set_index("hour", drop=True, inplace=True)
        return _meteo.to_dict(orient="index")

    @staticmethod
    def __theoretical_irrad(sunModel, day, skyType=PerrinDeBrichambaut.STANDARD_SKY):
        result = {}
        dt0 = datetime(*day.timetuple()[0:3], 0, tzinfo=timezone.utc)
        for h in range(24):
            dt = dt0 + timedelta(hours=h)
            alti, azim = sunModel.getSolarAnglesInDegrees(dt)
            sbdir = sunModel.getRadiationDirection(dt)
            dni, dhi, ghi = 0, 0, 0
            if (0.0 < alti):
                dni = PerrinDeBrichambaut.directNormalIrradiance(
                    AngleLib.toRadians(alti), skyType=skyType)
                dhi = PerrinDeBrichambaut.diffuseSolarIrradiance(
                    AngleLib.toRadians(alti), skyType)
                # GHI = DNI x cos(theta) + DHI
                ghi = dni * cos(AngleLib.toRadians(90-alti)) + dhi

            result[h] = {
                "sun_beam_dir": sbdir,
                "solar_alti": alti,
                "solar_azim": azim,
                "GHI": ghi,
                "DNI": dni,
                "DHI": dhi
            }
        return result

    @staticmethod
    def __in_direct_sunlight(angles, sunAlti, sunAzim):
        nangles = len(angles)
        offset = 360 / nangles
        position_in_vect = int((sunAzim + offset/2) / offset) % 64
        return 1 if (sunAlti > angles[position_in_vect]) else 0

    def runWithArgs(self, row):
        normal = (0, 0, 1)

        svf, angles = row[self.svf], row[self.angles]

        sw_direct, sw_diffuse = 0, 0
        beInTheSun = []

        for dt in self.dts:
            sunBeamDir, alti, azim, ghi, dni, dhi = self.irrad[dt.hour].values(
            )

            curr_beInTheSun = SkyMapRadiationBalance.__in_direct_sunlight(
                angles, alti, azim)
            beInTheSun.append(curr_beInTheSun)

            curr_sw_direct = curr_beInTheSun * dni * dot(sunBeamDir, normal)
            sw_direct += curr_sw_direct

            curr_sw_diffuse = svf * dhi
            sw_diffuse += curr_sw_diffuse

        hours_in_sunlight = sum(beInTheSun)
        hours_of_shade = len(beInTheSun) - hours_in_sunlight
        if self.encode:
            beInTheSun = ArrayCoding.encode(beInTheSun)
        return {
            "in_the_sun": beInTheSun,
            "hrs_in_sun": hours_in_sunlight,
            "hrs_shade": hours_of_shade,
            "sw_direct": sw_direct,
            "sw_diffuse": sw_diffuse,
            # "lw_atmos": lw_atmos,
            # "sum_sw_lw": sum_sw_lw,
        }


"""
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.morph.STSkyMap25D import STSkyMap25D
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess
from t4gpd.wrf.MeteoFrancePredictor import MeteoFrancePredictor
from t4gpd.wrf.MeteoFranceReader import MeteoFranceReader

buildings = GeoDataFrameDemos.ensaNantesBuildings()
sensors = STGrid(buildings, dx=80, dy=None, indoor=False,
    intoPoint=True, withDist=False).run() # < 0.1 sec.
skymaps = STSkyMap25D(buildings, sensors, nRays=64, rayLength=100,
	elevationFieldname="HAUTEUR", h0=0.0, size=1.0,
	withIndices=True, withAngles=True, encode=False).execute() # < 10 sec.

PERIODS = [
    # (2020, 2022),
    (2023, 2024),
]

rootdir = "/home/tleduc/data/meteofrance/data"
ifiles = [
    f"{rootdir}/HOR_departement_44_periode_{year0}-{year1}.csv"
    for year0, year1 in PERIODS
]
meteo = MeteoFranceReader(ifiles, station="NANTES-BOUGUENAIS").execute()
meteo2 = MeteoFrancePredictor(meteo).execute()

day = date(2023, 6, 21)
op = SkyMapRadiationBalance(skymaps, day, meteo=meteo2)
irrad1 = STGeoProcess(op, skymaps).run()

op = SkyMapRadiationBalance(skymaps, day, meteo=meteo)
irrad2 = STGeoProcess(op, skymaps).run()
"""

"""
import matplotlib.pyplot as plt
from pandas import read_csv, to_datetime
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.morph.STSkyMap25D import STSkyMap25D
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess

buildings = GeoDataFrameDemos.ensaNantesBuildings()
sensors = STGrid(buildings, dx=80, dy=None, indoor=False,
                 intoPoint=True, withDist=False).run()  # < 0.1 sec.
skymaps = STSkyMap25D(buildings, sensors, nRays=64, rayLength=100,
                      elevationFieldname="HAUTEUR", h0=0.0, size=9.0,
                      withIndices=True, withAngles=True, encode=False).run()  # < 1 sec.

meteo = read_csv(
    "/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology_extreme.csv")
meteo.timestamp = meteo.timestamp.apply(lambda dt: to_datetime(dt))
meteo.sun_beam_dir = meteo.sun_beam_dir.apply(lambda dt: ArrayCoding.decode(dt))

day = date(2022, 12, 22)
day = date(2022, 6, 13)
op = SkyMapRadiationBalance(skymaps, day, meteo=meteo)
irrad = STGeoProcess(op, skymaps).run()

fig, ax = plt.subplots(figsize=(1.0*8.26, 1.0*8.26))
buildings.plot(ax=ax)
# skymaps.plot(ax=ax, column="svf", legend=True, legend_kwds={"label": "svf"})
irrad.set_geometry("viewpoint").plot(ax=ax, column="hrs_in_sun", legend=True)
fig.tight_layout()
ax.axis("off")
plt.show()
"""

"""
import matplotlib.pyplot as plt
from pandas import read_csv, to_datetime
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.morph.STSkyMap25D import STSkyMap25D
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess

buildings = GeoDataFrameDemos.ensaNantesBuildings()
sensors = STGrid(buildings, dx=80, dy=None, indoor=False,
                 intoPoint=True, withDist=False).run()  # < 0.1 sec.
skymaps = STSkyMap25D(buildings, sensors, nRays=64, rayLength=100,
                      elevationFieldname="HAUTEUR", h0=0.0, size=9.0,
                      withIndices=True, withAngles=True, encode=False).run()  # < 1 sec.

meteo = read_csv(
    "/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology.csv")
meteo.sun_beam_dir = meteo.sun_beam_dir.apply(lambda dt: ArrayCoding.decode(dt))

day = date(2024, 3, 1)  # Focus on March
op = SkyMapRadiationBalance(skymaps, day, meteo=meteo)
irrad = STGeoProcess(op, skymaps).run()

fig, ax = plt.subplots(figsize=(1.0*8.26, 1.0*8.26))
buildings.plot(ax=ax)
# skymaps.plot(ax=ax, column="svf", legend=True, legend_kwds={"label": "svf"})
irrad.set_geometry("viewpoint").plot(ax=ax, column="sw_direct", legend=True)
fig.tight_layout()
ax.axis("off")
plt.show()
"""

"""
import matplotlib.pyplot as plt
from pandas import read_csv, to_datetime
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.morph.STSkyMap25D import STSkyMap25D
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess

buildings = GeoDataFrameDemos.ensaNantesBuildings()
sensors = STGrid(buildings, dx=80, dy=None, indoor=False,
                 intoPoint=True, withDist=False).run()  # < 0.1 sec.
skymaps = STSkyMap25D(buildings, sensors, nRays=64, rayLength=100,
                      elevationFieldname="HAUTEUR", h0=0.0, size=9.0,
                      withIndices=True, withAngles=True, encode=False).run()  # < 1 sec.

day = date(2022, 6, 13)  # Focus on the hottest day
# day = date(2022, 12, 22)  # Focus on the coldest day
# irrad = SkyMapRadiationBalance2(skymaps, day, meteo=None).run()

meteo = read_csv(
    "/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology_extreme.csv")
meteo.timestamp = meteo.timestamp.apply(lambda dt: to_datetime(dt))

# meteo = read_csv("/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology.csv")
meteo.sun_beam_dir = meteo.sun_beam_dir.apply(lambda dt: ArrayCoding.decode(dt))

op = SkyMapRadiationBalance(skymaps, day, meteo=meteo)
irrad1 = STGeoProcess(op, skymaps).run()
print(irrad1[["gid", "svf", "in_the_sun", "hrs_in_sun", "hrs_shade", "sw_direct", "sw_diffuse"]])

fig, ax = plt.subplots(figsize=(1.0*8.26, 1.0*8.26))
buildings.plot(ax=ax)
irrad1.set_geometry("viewpoint").plot(ax=ax, column="sw_direct", legend=True)
irrad1.set_geometry("viewpoint").apply(lambda x: ax.annotate(
	text=x.gid, xy=x.geometry.centroid.coords[0],
	color="black", size=12, ha="center"), axis=1)
fig.tight_layout()
ax.axis("off")
plt.show()
plt.close(fig)
"""



'''
Created on 25 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from copy import deepcopy
from geopandas import GeoDataFrame, sjoin_nearest
from networkx import betweenness_centrality, closeness_centrality, degree_centrality, DiGraph, NetworkXNoPath, shortest_path, to_numpy_array
from scipy.sparse.csgraph import minimum_spanning_tree
from shapely import delaunay_triangles, LineString, MultiPoint, Point
from shapely.ops import nearest_points
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraphVertex import UrbanGraphVertex


class UrbanGraph(object):
    '''
    classdocs
    '''

    def __init__(self, ciVertices=None, icVertices=None, edges=None,
                 nx_graph=None, dk_graph=None, gdfOfBipoints=None, crs=None):
        '''
        Constructor
        '''
        if not ((ciVertices is None) or (icVertices is None) or
                (nx_graph is None) or (gdfOfBipoints is None)):
            self.method = "networkx"

        elif not ((ciVertices is None) or (icVertices is None) or
                  (edges is None) or (dk_graph is None)):
            self.method = "dijkstar"

        elif not ((ciVertices is None) or (icVertices is None) or
                  (edges is None)):
            self.method = None

        else:
            raise NotImplementedError(
                "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

        self.ciVertices = ciVertices
        self.icVertices = icVertices
        self.edges = edges
        self.nx_graph = nx_graph
        self.dk_graph = dk_graph
        self.gdfOfBipoints = gdfOfBipoints
        self.crs = crs

    def __str__(self) -> str:
        if self.method is None:
            edges = ""
            for k1, v1 in self.edges.items():
                for k2, v2 in v1.items():
                    edges = f"{edges}\n{k1}->{k2}: {v2}"
            return f"""
{self.icVertices}
{edges}
"""

    def __get_next_vertex(self, prev, curr):
        succ = list(self.edges[curr])
        if 2 == len(succ):
            if (prev == succ[0]) and not (prev == succ[1]):
                return succ[1]
            elif (prev == succ[1]) and not (prev == succ[0]):
                return succ[0]
        return None

    def __burn_edge(self, idx1, idx2, burnedEdges):
        minIdx = min(idx1, idx2)
        maxidx = max(idx1, idx2)
        if minIdx in burnedEdges:
            burnedEdges[minIdx].add(maxidx)
        else:
            burnedEdges[minIdx] = set({maxidx})

    def __build_road_section(self, vertexIndex, nextVertexIndex, burnedEdges):
        result = []

        prevIdx = vertexIndex
        result.append(self.icVertices[prevIdx])
        currIdx = nextVertexIndex
        result.append(self.icVertices[currIdx])
        self.__burn_edge(prevIdx, currIdx, burnedEdges)

        nextIdx = self.__get_next_vertex(prevIdx, currIdx)
        while nextIdx is not None:
            result.append(self.icVertices[nextIdx])
            prevIdx = currIdx
            currIdx = nextIdx
            self.__burn_edge(prevIdx, currIdx, burnedEdges)

            if (nextIdx == vertexIndex):
                # current road section is a loop, stop the process
                nextIdx = None
            else:
                nextIdx = self.__get_next_vertex(prevIdx, currIdx)

        return LineString(result)

    def getUniqueRoadsSections(self):
        if (self.method in [None, "dijkstar"]):
            rows = []
            burnedEdges = dict()
            for vertexIndex in list(self.edges.keys()):
                nbConnectedVertices = len(self.edges[vertexIndex])
                if (2 != nbConnectedVertices):
                    # current vertex is a "boundary" of a "road section"
                    for nextVertexIndex in self.edges[vertexIndex]:
                        minIdx = min(vertexIndex, nextVertexIndex)
                        maxIdx = max(vertexIndex, nextVertexIndex)
                        if not (minIdx in burnedEdges and (maxIdx in burnedEdges[minIdx])):
                            geom = self.__build_road_section(
                                vertexIndex, nextVertexIndex, burnedEdges)
                            rows.append({
                                "geometry": geom,
                                "gid": len(rows),
                                "distance": geom.length
                            })
            return GeoDataFrame(rows, crs=self.crs)
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def getUniqueRoadsSectionsNodes(self):
        if (self.method in [None, "dijkstar"]):
            rows = []
            for k, v in self.edges.items():
                nbConnectedVertices = len(v)
                if (2 != nbConnectedVertices):
                    rows.append({
                        "gid": k,
                        "valency": nbConnectedVertices,
                        # "nb_connections": nbConnectedVertices,
                        "geometry": Point(self.icVertices[k]),
                    })
            return GeoDataFrame(rows, crs=self.crs)
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def getUniqueRoadsSectionsWithoutCulDeSac(self):
        rs = self.getUniqueRoadsSections()
        rows = []
        for _, row in rs.iterrows():
            _coords = list(row["geometry"].coords)
            startHash = UrbanGraphVertex.hash_coord(_coords[0])
            stopHash = UrbanGraphVertex.hash_coord(_coords[-1])
            if ((1 < len(self.edges[self.ciVertices[startHash]])) and
                    (1 < len(self.edges[self.ciVertices[stopHash]]))):
                rows.append(row)
        return GeoDataFrame(rows, crs=self.crs)

    def __fromGraphToGeoDataFrame1(self, nx_graph, centralities, label):
        rows = []
        for nodeIndex, centrality in centralities.items():
            rows.append({
                "gid": nodeIndex,
                "valency": len(nx_graph.adj[nodeIndex]),
                label: centrality,
                "geometry": Point((self.icVertices[nodeIndex])),
            })
        return GeoDataFrame(rows, crs=self.crs)

    def betweenness_centrality(self):
        if ("networkx" == self.method):
            centralities = betweenness_centrality(
                self.nx_graph, weight="weight")
            return self.__fromGraphToGeoDataFrame1(self.nx_graph, centralities, "betweenness")
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def closeness_centrality(self):
        if ("networkx" == self.method):
            centralities = closeness_centrality(
                self.nx_graph, distance="weight")
            return self.__fromGraphToGeoDataFrame1(self.nx_graph, centralities, "closeness")
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def degree_centrality(self):
        if ("networkx" == self.method):
            centralities = degree_centrality(self.nx_graph)
            return self.__fromGraphToGeoDataFrame1(self.nx_graph, centralities, "degree_c")
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def __fromGraphToGeoDataFrame2(self, nx_graph):
        rows = []
        for nodeIndex1, v1 in nx_graph.adj.items():
            for nodeIndex2, v2 in v1.items():
                rows.append({
                    "from": nodeIndex1,
                    "to": nodeIndex2,
                    "weight": v2["weight"],
                    "geometry": LineString([self.icVertices[nodeIndex1], self.icVertices[nodeIndex2]]),
                })
        return GeoDataFrame(rows, crs=self.crs)

    def minimum_spanning_tree(self):
        if ("networkx" == self.method):
            matrix = to_numpy_array(self.nx_graph)
            mst = minimum_spanning_tree(matrix)
            mst = DiGraph(mst.toarray())
            return self.__fromGraphToGeoDataFrame2(mst)
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def delaunay_triangulation(self):
        if (self.method in [None, "dijkstar"]):
            mpt = MultiPoint(list(self.icVertices.values()))
            tins = delaunay_triangles(mpt, tolerance=0, only_edges=True)
            rows = []
            for gid, linestring in enumerate(tins.geoms, start=1):
                rows.append({
                    "gid": gid,
                    "geometry": linestring,
                })
            return GeoDataFrame(rows, crs=self.crs)
        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")

    def __spaths_to_lineString(self, spaths):
        if 2 <= len(spaths):
            return LineString([
                self.icVertices[nodeIndex] for nodeIndex in spaths
            ])
        return None

    def __fromGraphToGeoDataFrame3(self, spaths, source, target):
        rows = []

        if (source is None) and (target is None):
            raise NotImplementedError("")

        elif (source is None) or (target is None):
            for spath in spaths.values():
                ls = self.__spaths_to_lineString(spath)
                if not ls is None:
                    rows.append({
                        "fromPoint": Point(self.icVertices[spath[0]]).wkt,
                        "toPoint": Point(self.icVertices[spath[-1]]).wkt,
                        "path": ArrayCoding.encode(spath),
                        "pathLen": ls.length,
                        "geometry": ls
                    })
        else:
            ls = self.__spaths_to_lineString(spaths)
            if not ls is None:
                rows.append({
                    "fromPoint": Point(self.icVertices[spaths[0]]).wkt,
                    "toPoint": Point(self.icVertices[spaths[-1]]).wkt,
                    "path": ArrayCoding.encode(spaths),
                    "pathLen": ls.length,
                    "geometry": ls
                })

        return GeoDataFrame(rows, crs=self.crs)

    def __add_two_new_edges(self, id1, id2):
        ls = LineString([self.icVertices[id1][0:2], self.icVertices[id2][0:2]])
        self.nx_graph.add_edge(id1, id2, weight=ls.length, object={
            "geometry": ls, "weight": ls.length})

        ls = LineString([self.icVertices[id2][0:2], self.icVertices[id1][0:2]])
        self.nx_graph.add_edge(id2, id1, weight=ls.length, object={
            "geometry": ls, "weight": ls.length})

    def _add_new_point(self, pt):
        if pt is None:
            return None
        if not isinstance(pt, Point):
            raise IllegalArgumentTypeException(pt, "Point or None")
        str_coord = UrbanGraphVertex.hash_coord(pt.coords[0])
        if str_coord in self.ciVertices:
            return self.ciVertices[str_coord]

        tmp = sjoin_nearest(GeoDataFrame([{"geometry": pt}], crs=self.crs),
                            self.gdfOfBipoints)

        if (0 < len(tmp)):
            # CAUTION WITH THE 2 FOLLOWING iat[]
            bipoint = self.gdfOfBipoints.iat[tmp.iat[0, 1], 1]
            _, rp = nearest_points(pt, bipoint)
            id1 = UrbanGraphVertex.add_vertex(
                self.ciVertices, self.icVertices, pt.coords[0])
            id2 = UrbanGraphVertex.add_vertex(
                self.ciVertices, self.icVertices, rp.coords[0])
            id3 = self.ciVertices[UrbanGraphVertex.hash_coord(
                bipoint.coords[0])]
            id4 = self.ciVertices[UrbanGraphVertex.hash_coord(
                bipoint.coords[-1])]
            self.__add_two_new_edges(id1, id2)
            self.__add_two_new_edges(id2, id3)
            self.__add_two_new_edges(id2, id4)

            return id1
        raise Exception("Unreachable instruction!")

    def shortest_path(self, sourcePt=None, targetPt=None):
        if ("networkx" == self.method):
            if not ((sourcePt is None) or (isinstance(sourcePt, Point))):
                raise IllegalArgumentTypeException(sourcePt, "Point or None")
            if not ((targetPt is None) or (isinstance(targetPt, Point))):
                raise IllegalArgumentTypeException(targetPt, "Point or None")
            if (sourcePt == targetPt):
                return GeoDataFrame([])

            ugClone = UrbanGraph(
                ciVertices=deepcopy(self.ciVertices),
                icVertices=deepcopy(self.icVertices),
                nx_graph=deepcopy(self.nx_graph),
                gdfOfBipoints=self.gdfOfBipoints.copy(deep=True),
                crs=self.crs)
            sourceIdx = ugClone._add_new_point(sourcePt)
            targetIdx = ugClone._add_new_point(targetPt)

            try:
                sp = shortest_path(ugClone.nx_graph, source=sourceIdx,
                                   target=targetIdx, weight="weight",
                                   method="dijkstra")
            except NetworkXNoPath as error:
                print(error)
                return GeoDataFrame()

            return ugClone.__fromGraphToGeoDataFrame3(sp, sourceIdx, targetIdx)

        raise NotImplementedError(
            "Bad invocation of the UrbanGraph constructor, use UrbanGraphFactory.create(...)")



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
# from shapely import minimum_rotated_rectangle
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.CaliperLib import CaliperLib


class RectangularityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, with_geom=False):
        self.with_geom = with_geom

    def runWithArgs(self, row):
        geom = row.geometry
        geomArea = geom.area
        geomPerim = geom.length

        # mabr = geom.minimum_rotated_rectangle
        mabr, len1, len2 = CaliperLib().mabr(geom)
        mabrArea = mabr.area
        mabrPerim = mabr.length

        stretching = min(len1, len2) / \
            max(len1, len2) if (0.0 < max(len1, len2)) else None
        areaRectDefect = geomArea / mabrArea if (0.0 < mabrArea) else None
        perimRectDefect = mabrPerim / geomPerim if (0.0 < geomPerim) else None

        if self.with_geom:
            return {
                "geometry": mabr,
                "stretching": stretching,
                "a_rect_def": areaRectDefect,
                "p_rect_def": perimRectDefect
            }

        return {
            "stretching": stretching,
            "a_rect_def": areaRectDefect,
            "p_rect_def": perimRectDefect
        }



'''
Created on 3 juil. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from geopandas import GeoDataFrame
from pandas import read_csv
from shapely import Point
from t4gpd.commons.GeoProcess import GeoProcess


class CSVReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, xFieldName="longitude", yFieldName="latitude",
                 fieldSep=",", srcEpsgCode="epsg:4326", dstEpsgCode=None, decimalSep="."):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.xFieldName = xFieldName
        self.yFieldName = yFieldName
        self.fieldSep = fieldSep
        self.crs = srcEpsgCode
        self.dstEpsgCode = dstEpsgCode
        self.decimalSep = decimalSep

    def run(self):
        df = read_csv(self.inputFile, sep=self.fieldSep,
                      decimal=self.decimalSep)
        df["geometry"] = df.apply(
            lambda row: Point(row[self.xFieldName], row[self.yFieldName]), axis=1)
        gdf = GeoDataFrame(df, crs=self.crs, geometry="geometry")
        if not self.dstEpsgCode is None:
            gdf = gdf.to_crs(self.dstEpsgCode)
        return gdf



'''
Created on 22 mar. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import asarray, cos, pi, sin, sqrt
from shapely import Point
from shapely.coords import CoordinateSequence
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.SphericalCartesianCoordinates import SphericalCartesianCoordinates


class DoubleProjectionLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def __to_pair_of_arrays(vp, rp):
        if (isinstance(vp, Point) and isinstance(rp, Point)):
            vp, rp = vp.coords, rp.coords
        if (isinstance(vp, CoordinateSequence) and isinstance(rp, CoordinateSequence)):
            vp, rp = vp[0], rp[0]
        vp, rp = asarray(vp), asarray(rp)
        return vp, rp

    @staticmethod
    def __first_projection(vp, rp):
        vp, rp = DoubleProjectionLib.__to_pair_of_arrays(vp, rp)

        vp2rp = rp - vp
        D = sqrt((vp2rp**2).sum())
        invD = 1 / D
        pp = invD * vp2rp

        return vp, pp, D

    @staticmethod
    def isoaire_projection(vp, rp, size=1):
        viewpoint, pp, D = DoubleProjectionLib.__first_projection(
            vp, rp)

        tmp = pp[0]**2 + pp[1]**2
        if (0 == tmp):
            # Projection of a point above the viewpoint
            return Point([
                viewpoint[0],
                viewpoint[1],
                D  # viewpoint[2]
            ])

        magn = size * sqrt((1-pp[2]) / tmp)
        return Point([
            viewpoint[0] + magn * pp[0],
            viewpoint[1] + magn * pp[1],
            D  # viewpoint[2]
        ])

    @staticmethod
    def orthogonal_projection(vp, rp, size=1):
        viewpoint, pp, D = DoubleProjectionLib.__first_projection(
            vp, rp)
        return Point([
            viewpoint[0] + size * pp[0],
            viewpoint[1] + size * pp[1],
            D  # viewpoint[2]
        ])

    @staticmethod
    def polar_projection(vp, rp, size=1):
        viewpoint, pp, D = DoubleProjectionLib.__first_projection(
            vp, rp)
        _, lon, lat = SphericalCartesianCoordinates.fromCartesianToSphericalCoordinates(
            *pp)
        magn = (size * (pi - 2 * lat)) / pi
        return Point([
            viewpoint[0] + magn * cos(lon),
            viewpoint[1] + magn * sin(lon),
            D  # viewpoint[2]
        ])

    @staticmethod
    def stereographic_projection(vp, rp, size=1):
        viewpoint, pp, D = DoubleProjectionLib.__first_projection(
            vp, rp)
        magn = size / (1.0 + pp[2])
        return Point([
            viewpoint[0] + magn * pp[0],
            viewpoint[1] + magn * pp[1],
            D  # viewpoint[2]
        ])

    @staticmethod
    def reverse_isoaire_projection(vp, pp, size=1):
        vp, pp = DoubleProjectionLib.__to_pair_of_arrays(vp, pp)
        pp = (pp - vp)/size
        tmp1 = 1 - pp[0]**2 - pp[1]**2
        tmp2 = sqrt(1 + tmp1)
        return Point([
            vp[0] + pp[0] * tmp2,
            vp[1] + pp[1] * tmp2,
            vp[2] + tmp1
        ])

    @staticmethod
    def reverse_orthogonal_projection(vp, pp, size=1):
        vp, pp = DoubleProjectionLib.__to_pair_of_arrays(vp, pp)
        pp = (pp - vp)/size
        return Point([
            vp[0] + pp[0],
            vp[1] + pp[1],
            vp[2] + sqrt(1.0 - pp[0]**2 - pp[1]**2)
        ])

    @staticmethod
    def reverse_polar_projection(vp, pp, size=1):
        raise NotImplementedError("Must be implemented!")

    @staticmethod
    def reverse_stereographic_projection(vp, pp, size=1):
        vp, pp = DoubleProjectionLib.__to_pair_of_arrays(vp, pp)
        pp = (pp - vp)/size
        magn = 2 / (1.0 + pp[0]**2 + pp[1]**2)
        return Point([
            vp[0] + magn * pp[0],
            vp[1] + magn * pp[1],
            vp[2] + magn - 1
        ])

    @ staticmethod
    def projection_switch(projectionName):
        projectionName = projectionName.lower()
        if ("stereographic" == projectionName):
            return DoubleProjectionLib.stereographic_projection
        elif ("orthogonal" == projectionName):
            return DoubleProjectionLib.orthogonal_projection
        elif ("polar" == projectionName):
            return DoubleProjectionLib.polar_projection
        elif ("isoaire" == projectionName):
            return DoubleProjectionLib.isoaire_projection
        raise IllegalArgumentTypeException(
            projectionName, "spherical projection as 'Stereographic', 'Orthogonal', 'Polar', or 'Isoaire'")

    @ staticmethod
    def reverse_projection_switch(projectionName):
        projectionName = projectionName.lower()
        if ("stereographic" == projectionName):
            return DoubleProjectionLib.reverse_stereographic_projection
        elif ("orthogonal" == projectionName):
            return DoubleProjectionLib.reverse_orthogonal_projection
        elif ("polar" == projectionName):
            return DoubleProjectionLib.reverse_polar_projection
        elif ("isoaire" == projectionName):
            return DoubleProjectionLib.reverse_isoaire_projection
        raise IllegalArgumentTypeException(
            projectionName, "spherical projection as 'Stereographic', 'Orthogonal', 'Polar', or 'Isoaire'")



'''
Created on 25 sept. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import MultiPoint
from shapely.ops import triangulate, voronoi_diagram
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STVoronoiPartition(object):
    '''
    classdocs
    '''

    def __init__(self, gdf):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        _multiPoints = []
        for geom in gdf.geometry:
            _multiPoints += GeomLib.getListOfShapelyPoints(geom, withoutClosingLoops=True)
        _multiPoints = MultiPoint(_multiPoints)

        self.gdf = gdf
        self.crs = gdf.crs
        self.multiPoints = _multiPoints

    def getDelaunayTriangles(self):
        return GeoDataFrame([{"geometry": t} for t in triangulate(self.multiPoints)], crs=self.crs)

    def run(self):
        vor = GeoDataFrame([{"geometry": v} for v in voronoi_diagram(
            self.multiPoints).geoms], crs=self.crs)
        vor = vor.sjoin(self.gdf, how="inner", predicate="intersects", rsuffix="right01234567889")
        vor.drop(columns="index_right01234567889", inplace=True)
        return vor



'''
Created on 1 juil. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, sjoin
from shapely.ops import unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STFacadesAnalysis import STFacadesAnalysis


class STDeploymentOfPVPanels(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, idFieldName='ID', elevationFieldName='HAUTEUR', coverRate=0.3,
                 azimuthRange=(270 - 45, 270 + 45), sizeOfPVPanel=(1.6, 1.0)):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'buildings GeoDataFrame')
        self.buildings = buildings

        if idFieldName not in buildings:
            raise Exception('%s is not a relevant field name!' % (idFieldName))
        self.idFieldName = idFieldName

        if elevationFieldName not in buildings:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))
        self.elevationFieldName = elevationFieldName

        self.coverRate = coverRate
        self.azim0, self.azim1 = azimuthRange
        self.dh, self.dw = sizeOfPVPanel

    def run(self):
        # ------
        # REMOVE SHARED WALLS
        uob = unary_union(self.buildings.geometry).buffer(1e-4)
        uob = GeoDataFrame([{'geometry': uob}], crs=self.buildings.crs)
        uob = uob.explode(ignore_index=True)

        # ------
        # SELECT WELL ORIENTED AND BIG ENOUGH FACADES
        facades = STFacadesAnalysis(uob, gidFieldname=None, elevationFieldname=None,
                                    exteriorOnly=False).run()
        facades = facades[(facades.azimuth >= self.azim0) & (facades.azimuth <= self.azim1) & 
                          (facades.length > (0.5 + min(self.dh, self.dw)))]
        facades.drop(columns=['buildingID', 'height', 'surface'], inplace=True)

        # ------
        # SPATIAL JOIN BETWEEN FACADES AND BUILDINGS
        _facades = facades.copy()
        _facades.geometry = _facades.geometry.apply(lambda g: g.centroid.buffer(1e-2))
        _facades = sjoin(_facades[['gid', 'geometry']],
                         self.buildings[[self.idFieldName, self.elevationFieldName, 'geometry']],
                         how='left', predicate='intersects')
        _facades.drop(columns=['index_right', 'geometry'], inplace=True)

        # ------
        # ATTRIBUTE JOIN BETWEEN FACADES AND _FACADES
        pvFacades = facades.merge(_facades, on='gid')

        # ------
        # NUMBER OF PHOTOVOLTAIC MODULES PER FACADES
        pvFacades['n_modules'] = list(zip(pvFacades.length, pvFacades.HAUTEUR))
        pvFacades.n_modules = pvFacades.n_modules.apply(lambda t: int(t[0] / self.dw) * int(t[1] / self.dh))
        pvFacades.n_modules = self.coverRate * pvFacades.n_modules  # apply the coverage rate
        pvFacades = pvFacades[pvFacades.n_modules > 1]  # Remove facades without modules
        pvFacades.n_modules = pvFacades.n_modules.astype(int)

        return pvFacades



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import pi, sqrt
# from shapely import minimum_bounding_circle
from t4gpd.commons.ChrystalAlgorithm import ChrystalAlgorithm
from t4gpd.commons.DiameterLib import DiameterLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class CircularityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, with_geom=False):
        self.with_geom = with_geom

    def runWithArgs(self, row):
        geom = row.geometry
        area, perim = geom.area, geom.length
        _, diamLen, _ = DiameterLib.diameter(geom)

        gravelius = perim / sqrt(4.0 * pi * area) if (0.0 < area) else None
        jaggedness = (perim * perim) / area if (0.0 < area) else None
        miller = (4.0 * pi * area) / (perim * perim) if (0.0 < perim) else None
        morton = (4.0 * area) / (pi * diamLen *
                                 diamLen) if (0.0 < diamLen) else None

        mbc, _, radius = ChrystalAlgorithm(geom).run()
        mbcArea = pi * radius * radius
        areaCircDefect = area / mbcArea if (0.0 < mbcArea) else None

        if self.with_geom:
            return {
                "geometry": mbc,
                "gravelius": gravelius,
                "jaggedness": jaggedness,
                "miller": miller,
                "morton": morton,
                "a_circ_def": areaCircDefect
            }

        return {
            "gravelius": gravelius,
            "jaggedness": jaggedness,
            "miller": miller,
            "morton": morton,
            "a_circ_def": areaCircDefect
        }



'''
Created on 30 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas import DataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from scipy.stats import kendalltau, pearsonr, spearmanr


class CorrelationLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def correlation_with_pvalues(df, method="pearson"):
        def __get_pvalues_dataframe(df, corr, method):
            from numpy import full, isnan, logical_or, nan, sum

            # Return a new array of given shape, filled with nan values
            pval = full(corr.shape, nan)
            for i in range(corr.shape[0]):
                for j in range(i, corr.shape[1]):
                    x = df.iloc[:, i]
                    y = df.iloc[:, j]
                    mask = ~logical_or(isnan(x), isnan(y))
                    if sum(mask) > 0:
                        # Get Pearson correlation p-value
                        pval[i, j] = pval[j, i] = method(x[mask], y[mask])[1]

            pval = DataFrame(pval, columns=corr.columns, index=corr.index)
            return pval

        if not isinstance(df, DataFrame):
            raise IllegalArgumentTypeException(df, "DataFrame")
        if "pearson" == method:
            _method = pearsonr
        elif "kendall" == method:
            _method = kendalltau
        elif "spearman" == method:
            _method = spearmanr
        else:
            raise IllegalArgumentTypeException(
                method, "pearson, kendall or spearman")

        _df = df.select_dtypes("number")
        corr = _df.corr(method=method)
        pval = __get_pvalues_dataframe(_df, corr, method=_method)

        return corr, pval

    @staticmethod
    def test():
        from numpy.random import default_rng

        nrows, rng = 10, default_rng(123)
        df = DataFrame(
            data=rng.uniform(low=0, high=1, size=(nrows, 4)),
            columns=["X1", "X2", "X3", "X4"])
        df = df.assign(
            Y1=lambda row: (10 * row.X1 + 0.5 * row.X2 + 0.4 * row.X3) / 10.9,
            Y2=lambda row: (10 * row.X1 + 1.5 * row.X2 + 1.4 * row.X4) / 12.9,
            Z=[f"z{i}" for i in range(len(df))]
        )

        corr, pval = CorrelationLib.correlation_with_pvalues(df)
        return df, corr, pval


# df, corr, pval = CorrelationLib.test()



'''
Created on 25 sep. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, ndarray, pi, sin, sqrt
from shapely import Point
from shapely.coords import CoordinateSequence
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AEProjectionLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def __to_point(vp):
        if isinstance(vp, (CoordinateSequence, list, ndarray, tuple)):
            vp = Point(vp)
        if not isinstance(vp, Point):
            raise IllegalArgumentTypeException(
                vp, "CoordinateSequence, list, ndarray, Point, or tuple")
        return vp

    @staticmethod
    def isoaire_projection(vp, azim, elev, size=1):
        # azim == lon, elev == lat
        vp = AEProjectionLib.__to_point(vp)
        radius = sqrt((1 - sin(elev)) / ((cos(elev)*cos(azim))
                      ** 2 + (cos(elev)*sin(azim))**2))
        radius *= (size * cos(elev))
        return (vp.x + radius * cos(azim), vp.y + radius * sin(azim))

    @staticmethod
    def orthogonal_projection(vp, azim, elev, size=1):
        # azim == lon, elev == lat
        vp = AEProjectionLib.__to_point(vp)
        radius = (size * cos(elev))
        return (vp.x + radius * cos(azim), vp.y + radius * sin(azim))

    @staticmethod
    def polar_projection(vp, azim, elev, size=1):
        # azim == lon, elev == lat
        vp = AEProjectionLib.__to_point(vp)
        radius = (size * (pi - 2 * elev)) / pi
        return (vp.x + radius * cos(azim), vp.y + radius * sin(azim))

    @staticmethod
    def stereographic_projection(vp, azim, elev, size=1):
        # azim == lon, elev == lat
        vp = AEProjectionLib.__to_point(vp)
        radius = (size * cos(elev)) / (1.0 + sin(elev))
        return (vp.x + radius * cos(azim), vp.y + radius * sin(azim))

    @staticmethod
    def reverse_isoaire_projection(vp, azim, elev, size=1):
        raise NotImplementedError("Must be implemented!")

    @staticmethod
    def reverse_orthogonal_projection(vp, pp, size=1):
        raise NotImplementedError("Must be implemented!")

    @staticmethod
    def reverse_polar_projection(vp, pp, size=1):
        raise NotImplementedError("Must be implemented!")

    @staticmethod
    def reverse_stereographic_projection(vp, pp, size=1):
        raise NotImplementedError("Must be implemented!")

    @ staticmethod
    def projection_switch(projectionName):
        projectionName = projectionName.lower()
        if ("stereographic" == projectionName):
            return AEProjectionLib.stereographic_projection
        elif ("orthogonal" == projectionName):
            return AEProjectionLib.orthogonal_projection
        elif ("polar" == projectionName):
            return AEProjectionLib.polar_projection
        elif ("isoaire" == projectionName):
            return AEProjectionLib.isoaire_projection
        raise IllegalArgumentTypeException(
            projectionName, "spherical projection as 'Stereographic', 'Orthogonal', 'Polar', or 'Isoaire'")

    @ staticmethod
    def reverse_projection_switch(projectionName):
        projectionName = projectionName.lower()
        if ("stereographic" == projectionName):
            return AEProjectionLib.reverse_stereographic_projection
        elif ("orthogonal" == projectionName):
            return AEProjectionLib.reverse_orthogonal_projection
        elif ("polar" == projectionName):
            return AEProjectionLib.reverse_polar_projection
        elif ("isoaire" == projectionName):
            return AEProjectionLib.reverse_isoaire_projection
        raise IllegalArgumentTypeException(
            projectionName, "spherical projection as 'Stereographic', 'Orthogonal', 'Polar', or 'Isoaire'")



'''
Created on 20 nov. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import asarray, isnan, pi
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.ellipse.EllipseLib import EllipseLib
from t4gpd.commons.ellipse.EllipticHullLib import EllipticHullLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class EllipticityAxisIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, threshold=None, debug=False, with_geom=False):
        '''
        Constructor
        '''
        self.mabe = EllipticHullLib(threshold, debug)
        self.with_geom = with_geom

    @staticmethod
    def __sort_of_linear_regression(mainAxis, pts):
        ppts = [mainAxis.interpolate(mainAxis.project(pt)) for pt in pts]
        dists = asarray([p.distance(pp) for p, pp in zip(pts, ppts)])
        mae, mse = dists.sum(), (dists ** 2).sum()
        return mae, mse

    def runWithArgs(self, row):
        geom = row.geometry

        pts = GeomLib.getListOfShapelyPoints(geom, withoutClosingLoops=True)
        chull = geom.convex_hull

        # [Point(centre), semiXAxis, semiYAxis, azim, nodes, methName]
        centre, semiXAxis, semiYAxis, azim, _, _ = self.mabe.getMinimumAreaBoundingEllipse(
            chull)

        geomArea = geom.area
        mabeArea = pi * semiXAxis * semiYAxis
        if (isnan(mabeArea)) or (mabeArea < geomArea):
            raise Exception(
                f"Minimum-area bounding ellipse is {mabeArea} (input geom area: {geomArea:.1f})!")

        mainAxis = EllipseLib.getEllipseMainAxis(
            centre, semiXAxis, semiYAxis, azim)
        flattening = min(semiXAxis, semiYAxis) / max(semiXAxis,
                                                     semiYAxis) if (0.0 < max(semiXAxis, semiYAxis)) else None
        mae, mse = EllipticityAxisIndices.__sort_of_linear_regression(
            mainAxis, pts)
        if self.with_geom:
            return {
                "geometry": mainAxis,
                "flattening": flattening,
                "ell_mae": mae,
                "ell_mse": mse
            }

        return {
            "flattening": flattening,
            "ell_mae": mae,
            "ell_mse": mse
        }



'''
Created on 6 juil. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re


class CSVLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def readLexeme(s, decimalSep='.'):
        if ('.' != decimalSep):
            s = s.replace(decimalSep, '.')
        try:
            nb = float(s)
            if nb.is_integer():
                return int(nb)
            return nb
        except ValueError:
            pass
        try:
            import unicodedata
            nb = unicodedata.numeric(s)
            if nb.is_integer():
                return int(nb)
            return nb
        except (TypeError, ValueError):
            pass
        return s



'''
Created on 3 dec. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from pandas import DatetimeIndex, concat
from pvlib.location import Location
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib


class PvlibProxy(object):
    '''
    classdocs
    '''
    @staticmethod
    def get_ghi_dni_dhi_positions(gdf, dtFieldname, altitude=0, model="ineichen"):
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        if not dtFieldname in gdf:
            raise IllegalArgumentTypeException(
                dtFieldname, "is not a valid timestamp field name")

        lat, lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)
        location = Location(lat, lon, altitude)

        dts = DatetimeIndex(gdf[dtFieldname])
        clear_sky = location.get_clearsky(dts, model=model)
        solar_position = location.get_solarposition(dts)

        result = concat([clear_sky, solar_position], axis=1)
        result.reset_index(names="timestamp", inplace=True)
        # result = result.loc[:, ["timestamp", "ghi", "dni", "dhi",
        #                         "apparent_elevation", "azimuth"]]

        return result

    @staticmethod
    def append_ghi_dni_dhi_positions(gdf, dtFieldname, altitude=0, model="ineichen"):
        clearsky = PvlibProxy.get_ghi_dni_dhi_positions(
            gdf, dtFieldname, altitude, model)
        result = gdf.merge(clearsky, how="inner", on=dtFieldname)
        return result


"""
from t4gpd.picoclim.PicopattReader import PicopattReader

inputFile = "/media/tleduc/disk_20241128/uclimweb/data/picopatt_nantes/picopatt_nantes*loire*20240919*1829.csv"
# inputFile = "/media/tleduc/disk_20241128/uclimweb/data/picopatt_montpellier/picopatt_montpellier*antigone*20240710*1412.csv"
df = PicopattReader(inputFile).run()
df2 = PvlibProxy.append_ghi_dni_dhi_positions(df, "timestamp")
print(df.head(3))
print(df2.head(3))
"""



'''
Created on 19 dec. 2021

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.isovist.ExactIsovistLib import ExactIsovistLib


class STExactIsovistField2D(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, viewpoints, rayLength=100.0, delta=10.0):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(
                buildings, 'buildings GeoDataFrame')
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(
                viewpoints, 'viewpoints GeoDataFrame')
        if not (buildings.crs == viewpoints.crs):
            raise Exception(
                'Illegal argument: buildings and viewpoints must share shames CRS!')

        self.buildings = buildings
        # CLEAN GEOMETRIES
        self.buildings.geometry = self.buildings.geometry.apply(
            lambda g: g.buffer(0))

        self.viewpoints = viewpoints
        self.rayLength = rayLength
        self.delta = delta

    def run(self):
        rows = []
        for _, row in self.viewpoints.iterrows():
            _vp = row.geometry.centroid
            # Update 21.08.2024
            _rayLength = row[self.rayLength] if (
                self.rayLength in self.viewpoints) else self.rayLength
            _isov = ExactIsovistLib.run(
                self.buildings, _vp, _rayLength, self.delta)

            del (_isov['nodes'])
            _isov['artif_hori'] = _isov['artif_hori'].wkt
            _isov['solid_surf'] = _isov['solid_surf'].wkt
            _isov['occlu_surf'] = _isov['occlu_surf'].wkt

            # Update 21.08.2024
            _newrow = dict(row)
            _newrow["viewpoint"] = _newrow["geometry"]
            _newrow.update(_isov)
            rows.append(_newrow)
            # rows.append(_isov)

        return GeoDataFrame(rows, crs=self.buildings.crs)



'''
Created on 30 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay
from numpy.random import default_rng
from pandas import DataFrame, merge
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class CartesianProductLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def product(df1, df2):
        if not isinstance(df1, DataFrame):
            raise IllegalArgumentTypeException(df1, "DataFrame")
        if not isinstance(df2, DataFrame):
            raise IllegalArgumentTypeException(df2, "DataFrame")

        _df1, _df2 = df1.copy(deep=True), df2.copy(deep=True)
        rng = default_rng()

        while True:
            fieldname = str(rng.integers(low=0, high=1e12, size=1)[0])
            if (not (fieldname in df1) and not (fieldname in df2)):
                _df1[fieldname] = 0
                _df2[fieldname] = 0
                result = merge(_df1, _df2, on=fieldname)
                result.drop(columns=[fieldname], inplace=True)
                return result

    @staticmethod
    def product_within_distance(gdf1, gdf2, distance):
        if not isinstance(gdf1, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf1, "GeoDataFrame")
        if not isinstance(gdf2, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf2, "GeoDataFrame")
        if not GeoDataFrameLib.shareTheSameCrs(gdf1, gdf2):
            raise Exception(
                "Illegal argument: gdf1 and gdf2 must share shames CRS!")
        result = CartesianProductLib.product(gdf1, gdf2)
        result = result.loc[result.apply(
            lambda row: row.geometry_x.distance(row.geometry_y), axis=1) <= distance]
        result.reset_index(drop=True, inplace=True)
        return result

    @staticmethod
    def product_within_distance2(gdf1, gdf2, distance):
        if not isinstance(gdf1, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf1, "GeoDataFrame")
        if not isinstance(gdf2, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf2, "GeoDataFrame")
        if not GeoDataFrameLib.shareTheSameCrs(gdf1, gdf2):
            raise Exception(
                "Illegal argument: gdf1 and gdf2 must share shames CRS!")

        _gdf1 = gdf1.copy(deep=True)
        _gdf1["geometry_x"] = _gdf1.geometry

        _gdf2 = gdf2.copy(deep=True)
        _gdf2["geometry_y"] = _gdf2.geometry
        _gdf2.geometry = _gdf2.geometry.apply(lambda g: g.buffer(distance))
        _gdf1 = overlay(_gdf1, _gdf2, how="intersection", keep_geom_type=True)

        args = {}
        for fieldname in gdf1.columns:
            if (fieldname in gdf2.columns):
                args.update({f"{fieldname}_1": f"{fieldname}_x",
                             f"{fieldname}_2": f"{fieldname}_y"})

        _gdf1.rename(columns=args, inplace=True)
        _gdf1.drop(columns="geometry", inplace=True)
        return _gdf1


"""
from t4gpd.morph.STGrid import STGrid
from shapely import box

gdf = GeoDataFrame([{"geometry": box(0, 0, 100, 100)}])
gdf1 = STGrid(gdf, dx=50, dy=None, indoor=None, intoPoint=True,
              encode=True, withDist=False).run()
gdf2 = STGrid(gdf, dx=2, dy=None, indoor=None, intoPoint=True,
              encode=True, withDist=False).run()
actual = CartesianProductLib.product_within_distance(gdf1, gdf2, distance=5)
# actual.rename(columns={"geometry_y": "geometry"}, inplace=True)

actual2 = CartesianProductLib.product_within_distance2(gdf1, gdf2, distance=5)
# actual2.rename(columns={"geometry_y": "geometry"}, inplace=True)

actual.to_csv("/tmp/1.csv", index=False, sep=";")
actual2.to_csv("/tmp/2.csv", index=False, sep=";")

import matplotlib.pyplot as plt
fig, ax = plt.subplots(figsize=(1*8.26, 1*8.26))
gdf1.plot(ax=ax, marker="v")
gdf2.plot(ax=ax, marker="+")
actual.plot(ax=ax, marker=".")
plt.show()
plt.close(fig)

import matplotlib.pyplot as plt
fig, ax = plt.subplots(figsize=(1*8.26, 1*8.26))
gdf1.plot(ax=ax, marker="v")
gdf2.plot(ax=ax, marker="+")
actual2.plot(ax=ax, marker=".")
plt.show()
plt.close(fig)
"""

"""
a = DataFrame(["a%d" % i for i in range(1, 4)], columns=["chpA"])
b = DataFrame(["b%d" % i for i in range(1, 3)], columns=["chpB"])
c = CartesianProductLib.product(a, b)
print(f"{a}\n{b}\n\n{c}")
"""



'''
Created on 23 aug. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, sjoin
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPolygonize import STPolygonize


class OverlapLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def overlap(gdf, fieldnames, patchid="__PATCH_ID__"):
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        if (fieldnames is None) or (0 == len(fieldnames)):
            raise IllegalArgumentTypeException(
                fieldnames, "non empty list of fieldnames")
            # raise IllegalArgumentTypeException(
            #     fieldnames, "non empty list of primary keys")

        for fieldname in fieldnames:
            if (fieldname not in gdf):
                raise IllegalArgumentTypeException(
                    fieldname, "fieldname of gdf")
            # if ((pk not in gdf) or (pk in gdf) and (not gdf[pk].is_unique)):
                # raise IllegalArgumentTypeException(pk, "primary key of gdf")

        if patchid in gdf:
            raise IllegalArgumentTypeException(
                patchid, "fieldname unknown to gdf")

        patches = STPolygonize(gdf, patchid=patchid).run()

        representative_points = patches.copy(deep=True)
        representative_points.geometry = representative_points.geometry.apply(
            lambda geom: geom.representative_point())

        result = sjoin(representative_points, gdf,
                       how="inner", predicate="within")

        result = result.groupby(by=patchid, as_index=False).agg(
            {fieldname: list for fieldname in fieldnames})
        if "noverlays" not in result:
            result["noverlays"] = result[fieldnames[0]].apply(lambda v: len(v))

        result = result.merge(patches, on=patchid)
        result = GeoDataFrame(result, crs=gdf.crs, geometry="geometry")
        return result

    @staticmethod
    def test():
        import matplotlib.pyplot as plt
        from datetime import datetime, timedelta, timezone
        from matplotlib.colors import ListedColormap
        from shapely import box

        dt0 = datetime(2024, 8, 23, 12, tzinfo=timezone.utc)
        gdf = GeoDataFrame([
            {"gid": 100+i, "geometry": box(0, 0, i, i),
             "timestamp": (dt0 + timedelta(hours=i)).time()}
            for i in range(1, 4)
        ])
        result = OverlapLib.overlap(
            gdf, ["gid", "timestamp"], patchid="patch_id")
        print(result)

        # MAPPING
        my_rgb_cmap = ListedColormap(["blue", "green", "red"])

        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(
            1.6 * 8.26, 0.8 * 8.26), squeeze=False)

        ax = axes[0, 0]
        gdf.plot(ax=ax, column="gid", alpha=0.3, cmap=my_rgb_cmap)
        ax.axis("off")

        ax = axes[0, 1]
        # gdf.plot(ax=ax, column="gid", alpha=0.3, cmap=my_rgb_cmap)
        HATCHES = ["o", ".", "x"]
        for i, patch_id in enumerate(result.patch_id.unique()):
            result[result.patch_id == patch_id].boundary.plot(
                ax=ax, hatch=HATCHES[i])
        ax.axis("off")

        plt.tight_layout()
        plt.show()
        plt.close(fig)

        return gdf, result


# igdf, ogdf = OverlapLib.test()



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, dtype, ndarray
from rasterio.io import DatasetReader
from rasterio.transform import Affine
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess


class RTFromArrayToRaster(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, array, rasterOrRoi, ndv=-9999, debug=False):
        '''
        Constructor
        '''
        if not isinstance(array, (list, ndarray, tuple)):
            raise IllegalArgumentTypeException(
                array, "list, numpy ndarray, or tuple")
        self.array = RTFromArrayToRaster.__asarray(array)
        if (1 == self.array.ndim):
            self.array = self.array.reshape(1, -1)
        if (2 != self.array.ndim):
            raise IllegalArgumentTypeException(
                array, "*1D* or *2D* list, numpy ndarray, or tuple")

        if not isinstance(rasterOrRoi, (DatasetReader, GeoDataFrame)):
            raise IllegalArgumentTypeException(
                rasterOrRoi, "DatasetReader or GeoDataFrame")
        self.rasterOrRoi = rasterOrRoi
        self.ndv = ndv
        self.debug = debug

    @staticmethod
    def __asarray(array):
        array = asarray(array)
        if (array.dtype == dtype(float)):
            return array.astype("float32")
        elif (array.dtype == dtype(int)):
            # Conversion to uint16, int16, int32, etc. is automatic
            return array.astype("uint8")
        raise IllegalArgumentTypeException(array, "Array of int or float values")

    def __extract_metadata_from_DatasetReader(self, raster):
        if (self.array.shape[0] != raster.height):
            raise Exception(
                f"array.shape[0] ({self.array.shape[0]}) <> raster.height ({raster.height})!")
        if (self.array.shape[1] != raster.width):
            raise Exception(
                f"array.shape[1] ({self.array.shape[1]}) <> raster.width ({raster.width})!")

        out_meta = raster.meta.copy()
        out_meta.update({
            "driver": "GTiff",
            "height": raster.height,
            "width": raster.width,
            "transform": raster.transform,
            "crs": raster.crs,
            "count": 1,
            "dtype": self.array.dtype,
            "nodata": self.ndv
        })
        return out_meta

    def __extract_metadata_from_GeoDataFrame(self, roi):
        minx, miny, maxx, maxy = roi.total_bounds
        nrows, ncols = self.array.shape

        xres = (maxx-minx) / ncols
        yres = (maxy-miny) / nrows

        transform = Affine.translation(minx, maxy) * Affine.scale(xres, -yres)

        out_meta = {
            "driver": "GTiff",
            "height": nrows,
            "width": ncols,
            "transform": transform,
            "crs": roi.crs,
            "count": 1,
            "dtype": self.array.dtype,
            "nodata": self.ndv,
            # "blockxsize": 1,
            # "blockysize": 1,
            # "compress": "lzw",
        }
        return out_meta

    def run(self):
        if isinstance(self.rasterOrRoi, DatasetReader):
            out_meta = self.__extract_metadata_from_DatasetReader(
                self.rasterOrRoi)

        elif isinstance(self.rasterOrRoi, GeoDataFrame):
            out_meta = self.__extract_metadata_from_GeoDataFrame(
                self.rasterOrRoi)

        #         print(f"""\theight / nrows = {out_meta['height']}
        # \twidth / ncols  = {out_meta['width']}
        # \tminx           = {out_meta['transform'][2]}
        # \tmaxy           = {out_meta['transform'][5]}
        # \txres           = {out_meta['transform'][0]}
        # \tyres           = {out_meta['transform'][4]}
        # \tndv            = {out_meta['nodata']}
        # """)
        result = self._write_and_load(self.array, out_meta, self.debug)
        return result


"""
import contextily as ctx
import matplotlib.pyplot as plt
from t4gpd.raster.RTToFile import RTToFile
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from numpy import arange

buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
roi = STBBox(buildings, buffDist=0).run()

buildings.to_file("/tmp/buildings.shp")
roi.to_file("/tmp/roi.shp")

array = arange(20).reshape(4, -1).astype(float)
array = arange(20).reshape(2, -1).astype(float)
raster1 = RTFromArrayToRaster(array, roi).run()
RTToFile(raster1, "/tmp/raster1.tif").run()

fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
buildings.boundary.plot(ax=ax, color="red", hatch=".")
ctx.add_basemap(ax, crs=buildings.crs, source="/tmp/raster1.tif")
ax.axis("off")
plt.tight_layout()
plt.show()
"""

"""
from geopandas import read_file
from t4gpd.raster.RTClip import RTClip
from t4gpd.raster.RTLoad import RTLoad
from t4gpd.raster.RTToFile import RTToFile
from t4gpd.raster.STRasterize import STRasterize

srcdir = "/home/tleduc/prj/uenv_aau_msc_course/tps"
dtmdir = "/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m";
raster = RTLoad(f"{dtmdir}/MNT_L93_0354_6690.asc").run()

#~ CROP RASTER FILE
roi = read_file(f"{srcdir}/data/roi.shp")
dtm = RTClip(raster, roi).run()
RTToFile(dtm, f"{srcdir}/data/dtm.tif").run()

#~ BUILDINGS LAYER RASTERIZING
buildings = read_file(f"{srcdir}/data/buildings.shp")
_dsm1 = STRasterize(buildings, dtm, attr="HAUTEUR").run()
RTToFile(_dsm1, f"{srcdir}/data/_dsm1.tif").run()

#~ STATUE LAYER RASTERIZING
statue = read_file(f"{srcdir}/data/statue.shp")
_dsm2 = STRasterize(statue, dtm, attr="HAUTEUR").run()
RTToFile(_dsm2, f"{srcdir}/data/_dsm2.tif").run()

#~ DSM1 + DSM2 => DSM12
dsm12 = _dsm1.read(1) + _dsm2.read(1)
dsm12 = RTFromArrayToRaster(dsm12, roi).run()
RTToFile(dsm12, f"{srcdir}/data/dsm12.tif").run()
"""

"""
import matplotlib.pyplot as plt
from geopandas import GeoDataFrame
from numpy import arange, flip
from shapely import box
from t4gpd.morph.STGrid import STGrid
from t4gpd.raster.RTToFile import RTToFile

roi = GeoDataFrame([{"geometry": box(0, 0, 50, 50)}])
grid = STGrid(roi, dx=10, dy=None, indoor=None, intoPoint=False,
              encode=False, withDist=False).run()

array = flip(arange(25).reshape(5,-1), axis=0)
raster = RTFromArrayToRaster(array, roi, ndv=0, debug=False).run()

# MAPPING - OK WITH QGIS
roi.to_file("/tmp/roi.shp")
grid.to_file("/tmp/grid.shp")
RTToFile(raster, "/tmp/raster.tif").run()

# MAPPING - BUG WITH IMSHOW
fig, ax = plt.subplots(figsize=(1.4 * 8.26, 1.0 * 8.26))
roi.boundary.plot(ax=ax, color="green")
grid.plot(ax=ax, column="gid", cmap="magma", legend=True)
ax.imshow(raster.read(1), cmap="magma")
ax.axis("off")
fig.tight_layout()
plt.show()
plt.close(fig)
"""



'''
Created on 30 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import ones_like, triu
from pandas import DataFrame
from t4gpd.commons.CorrelationLib import CorrelationLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class PlotCorrelationMatrix(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, df, tri=True, pval=True, figsize=None,
                 method="pearson", ofile=None):
        '''
        Constructor
        '''
        if not isinstance(df, DataFrame):
            raise IllegalArgumentTypeException(df, "DataFrame")
        self.df = df
        self.tri = tri
        self.pval = pval
        self.figsize = (2.5*8.26, 2.2*8.26) if figsize is None else figsize
        self.method = method
        self.ofile = ofile

    @staticmethod
    def __assign_pvalue_annotations(ax, corr, pval, tri):
        from numpy import isnan, max, min

        # ones_like(): return an array of ones (i.e. True) with the
        # same shape and type as a given array
        # triu(): return a copy of an array with the elements below
        # the k-th diagonal zeroed
        mask_pvalues = triu(ones_like(pval), k=1)

        max_corr, min_corr = max(corr.max()), min(corr.min())

        for i in range(pval.shape[0]):
            for j in range(pval.shape[1]):
                if mask_pvalues[i, j]:
                    p_value = pval.iloc[i, j]
                    if not isnan(p_value):
                        corr_value = corr.iloc[i, j]
                        text_color = "black" if ((min_corr + 0.36) <= corr_value <= (
                            max_corr - 0.4)) else "white"

                        label = f"(p = {p_value:.2f})"
                        if p_value <= 0.01:
                            # include double asterisks for p-value <= 0.01
                            label += "**"
                        elif p_value <= 0.05:
                            # include single asterisk for p-value <= 0.05
                            label += "*"
                        ax.text(i + 0.5, j + 0.8, label,
                                ha="center", va="center",
                                fontsize=8, color=text_color)
                        if not tri:
                            ax.text(j + 0.5, i + 0.8, label,
                                    ha="center", va="center",
                                    fontsize=8, color=text_color)

    def __plot(self, corr, pval):
        import matplotlib.pyplot as plt
        from seaborn import heatmap

        if self.tri:
            # ones_like(): return an array of ones (i.e. True) with the
            # same shape and type as a given array
            # triu(): return a copy of an array with the elements below
            # the k-th diagonal zeroed
            mask = triu(ones_like(corr, dtype=bool), k=1)
        else:
            mask = None

        fig, ax = plt.subplots(figsize=self.figsize)

        # Create the Correlation Heatmap
        # Data will not be shown in cells where mask is True
        heatmap = heatmap(corr,
                          annot=True,
                          annot_kws={"fontsize": 10},
                          fmt=".2f",
                          linewidths=0.5,
                          cmap="RdBu",
                          mask=mask,
                          ax=ax)
        # Assign p-value annotations with asterisks for significance
        if self.pval:
            PlotCorrelationMatrix.__assign_pvalue_annotations(
                ax, corr, pval, self.tri)

        if self.ofile is None:
            plt.show()
        else:
            plt.savefig(self.ofile, bbox_inches="tight")
        plt.close(fig)

    def run(self):
        # https://tosinharold.medium.com/enhancing-correlation-matrix-heatmap-plots-with-p-values-in-python-41bac6a7fd77
        corr, pval = CorrelationLib.correlation_with_pvalues(
            self.df, method=self.method)

        self.__plot(corr, pval)
        return corr, pval

    @staticmethod
    def test():
        from numpy.random import default_rng

        nrows, rng = 10, default_rng(123)
        df = DataFrame(
            data=rng.uniform(low=0, high=1, size=(nrows, 4)),
            columns=["X1", "X2", "X3", "X4"])
        df = df.assign(
            Y1=lambda row: (10 * row.X1 + 0.5 * row.X2 + 0.4 * row.X3) / 10.9,
            Y2=lambda row: (10 * row.X1 + 1.5 * row.X2 + 1.4 * row.X4) / 12.9,
        )

        PlotCorrelationMatrix(df, tri=True, ofile=None).run()
        return df


# df = PlotCorrelationMatrix.test()



'''
Created on 19 juin 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, degrees, pi, radians, r_


class AngleLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def angleBetween(direction1, direction2):
        angle = AngleLib.normAzimuth(
            direction2) - AngleLib.normAzimuth(direction1)
        return angle % (2 * pi)

    @staticmethod
    def angleBetweenNodes(node1, orig, node2):
        return AngleLib.angleBetween([node1[0] - orig[0], node1[1] - orig[1]], [node2[0] - orig[0], node2[1] - orig[1]])

    @staticmethod
    def eastCCW2northCW(angle, degree=True):
        return AngleLib.northCW2eastCCW(angle, degree)

    @staticmethod
    def fromEastCCWAzimuthToOppositeSliceIds(azim, nslices, offset=0, degree=True):
        quadrant = nslices // 4
        sliceId = AngleLib.fromEastCCWAzimuthToSliceId(azim, nslices, degree)
        start = (sliceId + quadrant + offset) % nslices
        stop = (sliceId + 3*quadrant - offset) % nslices
        if (start < stop):
            return slice(start, stop + 1)
        return r_[slice(start, nslices), slice(0, stop + 1)]

    @staticmethod
    def fromEastCCWAzimuthToSliceId(azim, nslices, degree=True):
        if not degree:
            azim = degrees(azim)
        azim = AngleLib.normalizeAngle(azim)
        offset = 360 / nslices
        return int((azim + offset/2) / offset) % nslices

    @staticmethod
    def normalizeAngle(angle, degree=True):
        modulo = 360 if degree else 2 * pi
        return angle % modulo

    @staticmethod
    def normAzimuth(direction):
        angle = arctan2(direction[1], direction[0])
        return angle % (2 * pi)

    @staticmethod
    def northCW2eastCCW(angle, degree=True):
        quarto = 90 if degree else pi / 2
        return AngleLib.normalizeAngle(5 * quarto - angle, degree)

    @staticmethod
    def southCCW2northCW(angle, degree=True):
        half = 180 if degree else pi
        return AngleLib.normalizeAngle(half - angle, degree)

    @staticmethod
    def toDegrees(angle):
        return degrees(angle)

    @staticmethod
    def toRadians(angle):
        return radians(angle)



'''
Created on 27 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''


class UrbanGraphVertex(object):
    '''
    classdocs
    '''

    @ staticmethod
    def hash_coord(coord):
        str_coord = f"{coord[0]}_{coord[1]}"
        return str_coord

    @staticmethod
    def add_vertex(ciVertices, icVertices, coord):
        str_coord = UrbanGraphVertex.hash_coord(coord)
        if not str_coord in ciVertices:
            nodeIndex = len(ciVertices)
            ciVertices[str_coord] = nodeIndex
            icVertices[nodeIndex] = coord
        else:
            nodeIndex = ciVertices[str_coord]
        return nodeIndex



'''
Created on 30 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, deg2rad, ndarray, r_
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.SVFLib import SVFLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class OrientedSVF(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, skymaps, anglesFieldname="angles", method="quadrants", svf=2018):
        '''
        Constructor
        '''
        if not isinstance(skymaps, GeoDataFrame):
            raise IllegalArgumentTypeException(skymaps, "GeoDataFrame")
        self.skymaps = skymaps

        if anglesFieldname not in skymaps:
            raise Exception(f"{anglesFieldname} is not a relevant field name!")
        self.anglesFieldname = anglesFieldname

        method = method.lower()
        if ("quadrants" == method):
            offset = 8
        elif ("octants" == method):
            offset = 16
        elif ("duodecants" == method):
            offset = 24
        else:
            raise IllegalArgumentTypeException(
                method, "'quadrants', 'octants', or 'duodecants'")

        for angles in self.skymaps[self.anglesFieldname]:
            if (isinstance(angles, (list, ndarray, tuple)) and (0 == len(angles) % offset)):
                n = len(angles) // offset
                break
            raise Exception(
                f"\n{angles}\n\nmust be a collection of angles whose length is a multiple of {offset}.")

        if (1981 == svf):
            print(
                f"SVF calculation method: Oke (1981)")
            self.svf = SVFLib.svfAngles1981
        else:
            print(
                f"SVF calculation method: Bernard et al. (2018)")
            self.svf = SVFLib.svfAngles2018

        if (8 == offset):
            self.quadrants = [
                slice(n, 3*n),  # north
                r_[slice(7*n, 8*n), slice(0, n)],  # east
                slice(5*n, 7*n),  # south
                slice(3*n, 5*n),  # west
            ]
            self.method = self.__quadrants
        elif (16 == offset):
            self.octants = [
                slice(3*n, 5*n),  # north
                slice(n, 3*n),  # northeast
                r_[slice(15*n, 16*n), slice(0, n)],  # east
                slice(13*n, 15*n),  # southeast
                slice(11*n, 13*n),  # south
                slice(9*n, 11*n),  # southwest
                slice(7*n, 9*n),  # west
                slice(5*n, 7*n),  # northwest
            ]
            self.method = self.__octants
        elif (24 == offset):
            self.duodecants = [
                slice(5*n, 7*n),  # north
                slice(3*n, 5*n),  # north-northeast
                slice(n, 3*n),  # east-northeast
                r_[slice(23*n, 24*n), slice(0, n)],  # east
                slice(21*n, 23*n),  # east-southeast
                slice(19*n, 21*n),  # south-southeast
                slice(17*n, 19*n),  # south
                slice(15*n, 17*n),  # south-southwest
                slice(13*n, 15*n),  # west-southwest
                slice(11*n, 13*n),  # west
                slice(9*n, 11*n),  # west-northwest
                slice(7*n, 9*n),  # north-northwest
            ]
            self.method = self.__duodecants
        else:
            raise Exception("Unreachable instruction!")

    def __quadrants(self, angles):
        north, east, south, west = self.quadrants
        north = self.svf(deg2rad(angles[north]))
        east = self.svf(deg2rad(angles[east]))
        south = self.svf(deg2rad(angles[south]))
        west = self.svf(deg2rad(angles[west]))
        return {
            "svf4_north": north,
            "svf4_east": east,
            "svf4_south": south,
            "svf4_west": west,
        }

    def __octants(self, angles):
        north, northeast, east, southeast, south, southwest, west, northwest = self.octants
        north = self.svf(deg2rad(angles[north]))
        northeast = self.svf(deg2rad(angles[northeast]))
        east = self.svf(deg2rad(angles[east]))
        southeast = self.svf(deg2rad(angles[southeast]))
        south = self.svf(deg2rad(angles[south]))
        southwest = self.svf(deg2rad(angles[southwest]))
        west = self.svf(deg2rad(angles[west]))
        northwest = self.svf(deg2rad(angles[northwest]))

        return {
            "svf8_north": north,
            "svf8_northeast": northeast,
            "svf8_east": east,
            "svf8_southeast": southeast,
            "svf8_south": south,
            "svf8_southwest": southwest,
            "svf8_west": west,
            "svf8_northwest": northwest,
        }

    def __duodecants(self, angles):
        n, nne, ene, e, ese, sse, s, ssw, wsw, w, wnw, nnw = self.duodecants
        n = self.svf(deg2rad(angles[n]))
        nne = self.svf(deg2rad(angles[nne]))
        ene = self.svf(deg2rad(angles[ene]))
        e = self.svf(deg2rad(angles[e]))
        ese = self.svf(deg2rad(angles[ese]))
        sse = self.svf(deg2rad(angles[sse]))
        s = self.svf(deg2rad(angles[s]))
        ssw = self.svf(deg2rad(angles[ssw]))
        wsw = self.svf(deg2rad(angles[wsw]))
        w = self.svf(deg2rad(angles[w]))
        wnw = self.svf(deg2rad(angles[wnw]))
        nnw = self.svf(deg2rad(angles[nnw]))

        return {
            "svf12_n": n,
            "svf12_nne": nne,
            "svf12_ene": ene,
            "svf12_e": e,
            "svf12_ese": ese,
            "svf12_sse": sse,
            "svf12_s": s,
            "svf12_ssw": ssw,
            "svf12_wsw": wsw,
            "svf12_w": w,
            "svf12_wnw": wnw,
            "svf12_nnw": nnw,
        }

    def runWithArgs(self, row):
        angles = asarray(row[self.anglesFieldname])
        if not isinstance(angles, (list, ndarray, tuple)):
            raise IllegalArgumentTypeException(
                angles, "list, ndarray or tuple")

        return self.method(angles)

    @staticmethod
    def test(method="octants"):
        import matplotlib.pyplot as plt
        from numpy import sqrt
        from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
        from t4gpd.morph.STGrid import STGrid
        from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess
        from t4gpd.morph.geoProcesses.Translation import Translation
        from t4gpd.skymap.STSkyMap25D import STSkyMap25D

        buildings = GeoDataFrameDemos.ensaNantesBuildings()
        viewpoints = STGrid(buildings, dx=80, indoor=False,
                            intoPoint=True).run()
        viewpoints.sort_values(by="gid", inplace=True)
        viewpoints.drop(
            columns=["row", "column", "neighbors4", "neighbors8", "indoor"], inplace=True)
        # viewpoints = viewpoints.loc[viewpoints[viewpoints.gid == 8].index]

        nRays = 48 # 96
        skymaps = STSkyMap25D(
            buildings, viewpoints, nRays=nRays, rayLength=100,
            elevationFieldname="HAUTEUR", h0=0, size=4,
            withIndices=True, withAngles=True, encode=False).run()

        op = OrientedSVF(skymaps, "angles", method=method, svf=2018)
        skymaps2 = STGeoProcess(op, skymaps).run()

        # PLOTTING
        vmin, vmax = 0, 1.0

        fig, ax = plt.subplots(figsize=(1.5*8.26, 1.15*8.26))
        buildings.plot(ax=ax, color="grey")
        viewpoints.plot(ax=ax, color="red", marker="P")
        skymaps.set_geometry("viewpoint").plot(
            ax=ax, alpha=0.5, column="svf", cmap="viridis",
            vmin=vmin, vmax=vmax, legend=True,
            legend_kwds={"label": "svf*"})
        skymaps.apply(lambda x: ax.annotate(
            text=f"{x['svf']:.2f}", xy=x.geometry.centroid.coords[0],
            color="red", size=12, ha="center"), axis=1)

        if ("quadrants" == method):
            dxy = 10
            trios = [
                ("svf4_north", 0, dxy), ("svf4_east", dxy, 0),
                ("svf4_south", 0, -dxy), ("svf4_west", -dxy, 0),
            ]
        elif ("octants" == method):
            dxy = 10
            trios = [
                ("svf8_north", 0, dxy), ("svf8_northeast", dxy, dxy),
                ("svf8_east", dxy, 0), ("svf8_southeast", dxy, -dxy),
                ("svf8_south", 0, -dxy), ("svf8_southwest", -dxy, -dxy),
                ("svf8_west", -dxy, 0), ("svf8_northwest", -dxy, dxy),
            ]
        elif ("duodecants" == method):
            dxy = 15
            da, db = 0.5 * dxy, (dxy * sqrt(3))/2
            trios = [
                ("svf12_n", 0, dxy), ("svf12_nne", da, db),
                ("svf12_ene", db, da), ("svf12_e", dxy, 0),
                ("svf12_ese", db, -da), ("svf12_sse", da, -db),
                ("svf12_s", 0, -dxy), ("svf12_ssw", -da, -db),
                ("svf12_wsw", -db, -da), ("svf12_w", -dxy, 0),
                ("svf12_wnw", -db, da), ("svf12_nnw", -da, db),
            ]
        for indic, dx, dy in trios:
            op = Translation((dx, dy))
            skymaps3 = STGeoProcess(op, skymaps2).run()
            skymaps3.plot(ax=ax, alpha=0.5, column=indic,
                          cmap="viridis", vmin=vmin,
                          vmax=vmax)
            skymaps3.apply(lambda x: ax.annotate(
                text=f"{x[indic]:.2f}", xy=x.geometry.centroid.coords[0],
                color="black", size=12, ha="center"), axis=1)
        ax.axis("off")
        fig.tight_layout()
        plt.show()
        plt.close(fig)
        return viewpoints, skymaps, skymaps2


# viewpoints, skymaps, skymaps2 = OrientedSVF.test()
